################################################################
#             Batch Execution of Kestrel Huntflow
#
# Usage: `python3 -m kestrel [-v] [--debug] hunting.hf`
################################################################

import argparse
import logging
import os

from kestrel.session import Session


def logging_setup(session, verbose_mode, debug_mode):
    # setup logging format, channel and granularity
    log_format = "%(asctime)s %(levelname)s %(name)s %(message)s"
    log_console = logging.StreamHandler()
    if session:
        log_file_path = os.path.join(session.runtime_directory, "session.log")
        log_file = logging.FileHandler(log_file_path)
        log_handlers = [log_console, log_file] if verbose_mode else [log_file]
    else:
        log_handlers = [log_console]
    logging.basicConfig(
        format=log_format,
        datefmt="%H:%M:%S",
        level=logging.DEBUG if debug_mode else logging.INFO,
        handlers=log_handlers,
    )


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Kestrel Interpreter")
    parser.add_argument("huntflow", help="huntflow in .hf file")
    parser.add_argument(
        "-v", "--verbose", help="print verbose log", action="store_true"
    )
    parser.add_argument(
        "--debug", help="debug level log (default is info level)", action="store_true"
    )
    args = parser.parse_args()

    logging_setup(None, args.verbose, args.debug)
    with Session(debug_mode=args.debug) as session:
        logging_setup(session, args.verbose, args.debug)
        with open(args.huntflow, "r") as fp:
            huntflow = fp.read()
        outputs = session.execute(huntflow)
        results = "\n\n".join([o.to_string() for o in outputs])
        print(results)



"""Docker analytics interface executes Kestrel analytics via docker.

An analytics using this interface should follow the rules:

- The analytics is built into a docker container reachable by the Python ``docker`` package.

- The name of the container should start with ``kestrel-analytics-``.

- The container will be launched with a mounted volumn ``/data/`` for exchanging input/output.

- The input Kestrel variables (all records) are put in ``/data/input/`` as
  ``0.parquet.gz``, ``1.parquet.gz``, ..., in the same order as they are passed
  to the ``APPLY`` command.

- The output (updated variable data) should be yielded by the analytics to
  ``/data/output/`` as ``0.parquet.gz``, ``1.parquet.gz``, ..., in the same
  order of the input variables. If a variable is unchanged, the output parquet
  file of it can be omitted.

- If a display object is yielded, the analytics should write it into
  ``/data/display/``.

"""

import docker
import logging
import pandas

from kestrel.analytics import AbstractAnalyticsInterface
from kestrel.exceptions import (
    InvalidAnalytics,
    AnalyticsManagerInternalError,
    AnalyticsError,
)
from kestrel.utils import mkdtemp
from kestrel_analytics_docker.config import (
    DOCKER_IMAGE_PREFIX,
    VAR_FILE_SUFFIX,
    INPUT_FOLDER,
    OUTPUT_FOLDER,
    DISP_FOLDER,
)
from kestrel.codegen.display import DisplayHtml

_logger = logging.getLogger(__name__)


class DockerInterface(AbstractAnalyticsInterface):
    @staticmethod
    def schemes():
        """Docker analytics interface only supports ``docker://`` scheme."""
        return ["docker"]

    @staticmethod
    def list_analytics():
        """Check docker for the list of Kestrel analytics."""
        docker_images = docker.from_env().images.list()
        tags = [tag for img in docker_images for tag in img.attrs["RepoTags"]]
        image_names = [tag.split(":")[0] for tag in tags]
        analytics_names = [
            img[len(DOCKER_IMAGE_PREFIX) :]
            for img in image_names
            if img.startswith(DOCKER_IMAGE_PREFIX)
        ]
        _logger.debug(f"Analytics names obtained: {str(analytics_names)}")
        return analytics_names

    @staticmethod
    def execute(uri, argument_variables, session_id=None, parameters=None):
        """Execute an analytics."""

        scheme, analytics_name = uri.split("://")
        container_name = DOCKER_IMAGE_PREFIX + analytics_name

        if scheme != "docker":
            raise AnalyticsManagerInternalError(
                f"interface {__package__} should not process scheme {scheme}"
            )

        shared_dir = mkdtemp().resolve()
        input_dir = shared_dir / INPUT_FOLDER
        input_dir.mkdir()
        output_dir = shared_dir / OUTPUT_FOLDER
        output_dir.mkdir()
        display_dir = shared_dir / DISP_FOLDER
        display_dir.mkdir()

        for index, var in enumerate(argument_variables):
            arg_var_file_name = str(index) + VAR_FILE_SUFFIX
            arg_var_file_path = str(input_dir / arg_var_file_name)
            var.dump_to_file(arg_var_file_path)

        dclient = docker.from_env()
        if not dclient.images.list(container_name):
            raise InvalidAnalytics(
                analytics_name,
                "docker",
                f"{container_name} is not an avaliable docker container.",
            )

        # the execution of the container
        try:
            dclient.containers.run(
                container_name,
                volumes={str(shared_dir): {"bind": "/data", "mode": "rw"}},
                environment=parameters,
            )
        except docker.errors.ContainerError as e:
            error = str(e)
            _logger.error(error)
            raise AnalyticsError(f"{analytics_name} failed: {error}")

        # process returned/updated variables
        for index, var in enumerate(argument_variables):
            arg_var_file_name = str(index) + VAR_FILE_SUFFIX
            arg_var_file_path = str(output_dir / arg_var_file_name)

            var_data = None
            try:
                var_data = pandas.read_parquet(arg_var_file_path)
            except FileNotFoundError:
                _logger.info(
                    f'analytics "{analytics_name}" has no output file {arg_var_file_name}.'
                )
            except Exception as e:
                raise AnalyticsError(
                    f'{e.__class__.__name__}: analytics "{analytics_name}" yielded invalid return data {arg_var_file_path}.'
                )

            if var_data is None:
                continue

            if not "id" in var_data:
                raise AnalyticsError(
                    f'analytics "{analytics_name}" yielded invalid return {arg_var_file_path} without "id".'
                )

            if not "type" in var_data:
                raise AnalyticsError(
                    f'analytics "{analytics_name}" yielded invalid return {arg_var_file_path} without "type".'
                )

            var_data_type_set = set(var_data["type"])
            var_data_type = var_data_type_set.pop()
            if var_data_type_set:
                raise AnalyticsError(
                    f'analytics "{analytics_name}" yielded invalid return {arg_var_file_path} with inconsistent types'
                )

            var_data_dict = var_data.to_dict(orient="records")
            var.store.reassign(var.entity_table, var_data_dict)

        # process returned display
        disp_files = list(display_dir.iterdir())
        if disp_files:
            disp_file = disp_files.pop()
            if disp_files:
                raise AnalyticsError(
                    f'analytics "{analytics_name}" yielded more than one display files'
                )
            if disp_file.suffix == ".html":
                with open(disp_file, "r") as h:
                    html = h.read()
                    display = DisplayHtml(html)
            else:
                raise NotImplementedError
        else:
            display = None

        return display



import json
import re
import uuid
import pathlib
import requests

from stix2matcher.matcher import match

from kestrel.datasource import AbstractDataSourceInterface
from kestrel.datasource import ReturnFromFile
from kestrel.exceptions import DataSourceManagerInternalError, DataSourceConnectionError


def _make_query_dir(uri):
    path = pathlib.Path(str(uuid.uuid5(uuid.NAMESPACE_URL, str(uri))))
    path.mkdir(parents=True, exist_ok=True)
    return path


def fixup_pattern(pattern):
    # The matcher doesn't accept TimestampLiterals in START/STOP
    # See https://github.com/oasis-open/cti-pattern-validator/issues/52
    return re.sub(r"(START|STOP)\s+t'", r"\1 '", pattern)


class StixBundleInterface(AbstractDataSourceInterface):
    @staticmethod
    def schemes():
        return ["file", "http", "https"]

    @staticmethod
    def list_data_sources():
        """Empty return for now.

        May not be meaningful to implement this for this interface. This method
        should enumerate hosts but the set is unbounded.

        """
        return []

    @staticmethod
    def query(uri, pattern, session_id=None):
        scheme, data_path = uri.split("://")
        pattern = fixup_pattern(pattern)

        ingestdir = _make_query_dir(uri)
        ingestfile = ingestdir / "data.json"

        # TODO: keep files in LRU cache?

        if scheme == "file":
            try:
                with open(data_path, "r") as f:
                    data = f.read()
            except Exception:
                raise DataSourceConnectionError(uri)
        elif scheme == "http" or scheme == "https":
            try:
                data = requests.get(uri)
            except requests.exceptions.ConnectionError:
                raise DataSourceConnectionError(uri)
        else:
            raise DataSourceManagerInternalError(
                f"interface {__package__} should not process scheme {scheme}"
            )

        bundle_in = json.loads(data)
        bundle_out = {}
        for prop, val in bundle_in.items():
            if prop == "objects":
                bundle_out[prop] = []
                for obj in val:
                    if obj["type"] != "observed-data" or match(pattern, [obj], False):
                        bundle_out[prop].append(obj)
            else:
                bundle_out[prop] = val

        with ingestfile.open("w") as f:
            json.dump(bundle_out, f)

        return ReturnFromFile(ingestdir.name, [str(ingestfile.resolve())])



import pathlib
import os
import uuid


KESTREL_CONFIG = pathlib.Path("kestrel") / "kestrel.toml"


def config_paths():
    # latter ones will override former ones
    paths = [
        # pip install with root
        pathlib.Path("/") / KESTREL_CONFIG,
        # pip install in venv
        pathlib.Path(os.getenv("VIRTUAL_ENV", "")) / "etc" / KESTREL_CONFIG,
        # conda environment install
        pathlib.Path(os.getenv("CONDA_PREFIX", "")) / "etc" / KESTREL_CONFIG,
        # GitHub Action: Ubuntu environment
        pathlib.Path(os.getenv("pythonLocation", "")) / "etc" / KESTREL_CONFIG,
        # pip install --user
        pathlib.Path(os.getenv("HOME", "")) / ".local" / "etc" / KESTREL_CONFIG,
        # user-defined configuration
        pathlib.Path(os.getenv("HOME", "")) / ".config" / KESTREL_CONFIG,
    ]

    path_dedups = []
    for path in paths:
        if path not in path_dedups:
            path_dedups.append(path)

    return path_dedups


def remove_empty_dicts(ds):
    # remove dict with all values as None in list({string:string})
    # this is the results from SQL query
    return [d for d in ds if set(d.values()) != {None}]


def dedup_dicts(ds):
    # deduplicate list({string:string})
    # this is the results from SQL query
    return [dict(s) for s in set(frozenset(d.items()) for d in ds)]


def dedup_ordered_dicts(ds):
    # deduplicate list({string:string})
    # maintain the order if seen
    res = []
    seen = set()
    for d in ds:
        s = str(d)
        if s not in seen:
            res.append(d)
        seen.add(s)
    return res


def subgroup_list(xs, gsize):
    return [xs[i : i + gsize] for i in range(0, len(xs), gsize)]


def mkdtemp():
    # create a temporary directory and return it (named after a random uuid)
    ps = None
    while not ps or pathlib.Path(ps).exists():
        ps = str(uuid.uuid4())
    p = pathlib.Path(ps)
    p.mkdir(parents=True, exist_ok=True)
    return p


class set_current_working_directory:
    def __init__(self, new_cwd):
        self.tmp_cwd = new_cwd

    def __enter__(self):
        self.cwd = os.getcwd()
        os.chdir(self.tmp_cwd)

    def __exit__(self, exception_type, exception_value, traceback):
        os.chdir(self.cwd)



from lark import Lark, Visitor, UnexpectedCharacters, UnexpectedToken
from collections import defaultdict

from kestrel.exceptions import InvalidStixPattern
from firepit.stix20 import get_grammar


def parse_extended_stix_pattern(pattern):
    grammar = get_grammar()
    try:
        ast = Lark(grammar, parser="lalr").parse(pattern)
    except UnexpectedCharacters as err:
        raise InvalidStixPattern(pattern, err.line, err.column, "character", err.char)
    except UnexpectedToken as err:
        raise InvalidStixPattern(pattern, err.line, err.column, "token", err.token)
    retor = ReferenceExtractor()
    retor.visit(ast)
    return retor.references


class ReferenceExtractor(Visitor):
    def __init__(self):
        self.references = defaultdict(list)
        super().__init__()

    def reference(self, tree):
        varname = tree.children[0].value
        attribute = tree.children[1].value
        self.references[varname].append(attribute)



import json
import pathlib
import uuid

import pandas as pd

from kestrel.exceptions import MissingEntityType, NonUniformEntityType


def load_data(
    store, output_entity_table, input_data, input_entity_type=None, query_id=None
):
    # if input data is list of strings, load directly
    # logic handled in store
    if type(input_data) == str:
        data = json.loads(input_data)
    else:
        data = input_data
    if type(data) == list and type(data[0]) == str:
        if not input_entity_type:
            raise MissingEntityType
        data = data
        entity_type = input_entity_type
    else:
        # load any table-like input data with header information
        dataframe = pd.DataFrame(data)
        entity_type = _extract_uniform_type(dataframe, input_entity_type)
        data = dataframe.to_dict(orient="records")

    store.load(output_entity_table, data, entity_type, query_id)
    return entity_type


def load_data_file(store, output_entity_table, file_path, input_entity_type=None):
    dump_format = _get_dump_format(file_path)
    if dump_format == "csv":
        data = pd.read_csv(file_path)
    elif dump_format == "parquet":
        data = pd.read_parquet(file_path)
    elif dump_format == "json":
        with open(file_path) as input_file:
            data = json.load(input_file)
    query_id = str(uuid.uuid5(uuid.NAMESPACE_URL, str(file_path)))
    entity_type = load_data(
        store, output_entity_table, data, input_entity_type, query_id
    )
    return entity_type


def dump_data_to_file(store, input_entity_table, file_path):
    p = pathlib.Path(file_path)
    p.parent.mkdir(parents=True, exist_ok=True)

    input_data = store.lookup(input_entity_table) if input_entity_table else []
    dump_format = _get_dump_format(p)
    if dump_format == "csv":
        pd.DataFrame(input_data).to_csv(file_path)
    elif dump_format == "parquet":
        pd.DataFrame(input_data).to_parquet(file_path)
    elif dump_format == "json":
        data = pd.DataFrame(input_data).to_json(orient="records")
        with open(file_path, "w") as output_file:
            output_file.write(data)


def _extract_uniform_type(dataframe, input_entity_type):
    # check if the input dataframe has "type" attribute and type is uniform
    # if passed, return entity type
    # if dataframe does not have "type", and input_entity_type is not None
    # use that, otherwise, raise Exception
    if "type" in dataframe:
        all_input_entity_types = dataframe["type"].unique()
        if len(all_input_entity_types) > 1:
            raise NonUniformEntityType(all_input_entity_types)
        else:
            return all_input_entity_types[0]
    elif input_entity_type:
        dataframe.loc[:, "type"] = input_entity_type
        return input_entity_type
    else:
        raise MissingEntityType


def _get_dump_format(path):
    # input: path is a path string or a Path from pathlib
    if isinstance(path, str):
        path = pathlib.Path(path)
    suffixes = path.suffixes
    if suffixes[-1] in [".gz", ".bz2", ".zip", ".xz"]:
        # supported compression suffix in pandas
        suffix = suffixes[-2]
    else:
        suffix = suffixes[-1]
    suffix = suffix[1:]
    if suffix in ["csv", "parquet", "json"]:
        return suffix
    else:
        raise NotImplementedError("unknown file dump format")



################################################################
#                     Generic Errors
################################################################


class KestrelException(Exception):
    """Generic Kestrel Exception

    Args:
        error (str): error message.
        suggestion (str): suggestion to fix the issue.
    """

    def __init__(self, error, suggestion=""):
        self.error = error if error[-1] == "." else error + "."
        self.suggestion = suggestion if suggestion[-1] == "." else suggestion + "."

    def __str__(self):
        return f"[ERROR] {self.__class__.__name__}: {self.error} {self.suggestion}"


class KestrelInternalError(KestrelException):
    def __init__(self, error):
        super().__init__(error, "please open a github issue to report")


################################################################
#                     Kestrel Session Errors
################################################################


class NoValidConfiguration(KestrelException):
    def __init__(self):
        super().__init__(
            "no valid configuration files found",
            'reinstall package with "pip install". report bug if not solved',
        )


################################################################
#                       Kestrel Syntax Errors
################################################################


class KestrelSyntaxError(KestrelException):
    def __init__(self, line, column, invalid_term_type, invalid_term_value):
        self.line = line
        self.column = column
        self.invalid_term_type = invalid_term_type
        self.invalid_term_value = invalid_term_value
        super().__init__(
            f'invalid {self.invalid_term_type} "{self.invalid_term_value}" at line {self.line} column {self.column}',
            "rewrite the failed statement",
        )


class InvalidStixPattern(KestrelException):
    def __init__(
        self,
        stix,
        line=None,
        column=None,
        invalid_term_type=None,
        invalid_term_value=None,
    ):
        self.stix = stix
        self.line = line
        self.column = column
        self.invalid_term_type = invalid_term_type
        self.invalid_term_value = invalid_term_value
        msg = f'invalid STIX pattern "{stix}"'
        if self.invalid_term_value:
            details = f': invalid {self.invalid_term_type} "{self.invalid_term_value}" at line {self.line} column {self.column}'
            msg = msg + details
        super().__init__(msg, "rewrite the STIX pattern")


class VariableNotExist(KestrelException):
    def __init__(self, var_name):
        super().__init__(
            f'variable "{var_name}" does not exist', "check the variable used"
        )


class UnsupportedRelation(KestrelException):
    def __init__(self, entity_x, relation, entity_y):
        super().__init__(
            f'unsupported relation "{entity_x}--{relation}--{entity_y}"',
            "check for supported relations and entity types in the documentation",
        )


class UnsupportedStixSyntax(KestrelException):
    def __init__(self, msg):
        super().__init__(msg, "rewrite the STIX pattern")


################################################################
#                 Kestrel Code Generation Errors
################################################################


class EmptyInputVariable(KestrelException):
    def __init__(self, var_name):
        super().__init__(
            f'empty input variable "{var_name}"',
            "rewrite the command to use a non-empty variable as input",
        )


class InvalidAttribute(KestrelException):
    def __init__(self, attribute):
        super().__init__(
            f'invalid attribute "{attribute}"',
            "rewrite the command with a valid attribute",
        )


class NonUniformEntityType(KestrelException):
    def __init__(self, etypes):
        super().__init__(
            f"there are more than one entity types in input data: {etypes}",
            "provide homogeneous entities to construct a Kestrel variable",
        )


class MissingEntityType(KestrelException):
    def __init__(self):
        super().__init__(
            'input data does not have "type" column',
            'add "type" column to data or specify entity type in the Kestrel command',
        )


################################################################
#                     Data Source Errors
################################################################


class DataSourceConnectionError(KestrelException):
    def __init__(self, uri):
        super().__init__(
            f"cannot establish connection to {uri}",
            "check URI for typos; please test network connection",
        )


class DataSourceManagerInternalError(KestrelInternalError):
    def __init__(self, error):
        super().__init__(error)


class InvalidDataSource(KestrelException):
    def __init__(self, uri, itf, msg=""):
        super().__init__(
            f'invalid data source "{uri}" at interface "{itf}". {msg}',
            "please check data source configuration",
        )


class DataSourceError(KestrelException):
    def __init__(self, error):
        super().__init__(
            f"data source internal error: {error}", "please test data source manually"
        )


class DataSourceInterfaceNotFound(KestrelException):
    def __init__(self, scheme):
        super().__init__(
            f'interface handling "{scheme}://" does not exist',
            "(re)install the missing/broken data source interface package",
        )


class InvalidDataSourceInterfaceImplementation(Exception):
    def __init__(self, error):
        super().__init__(error, "report to data source interface developer")


class ConflictingDataSourceInterfaceScheme(KestrelException):
    def __init__(self, itf_a, itf_b, scheme):
        super().__init__(
            f'conflicting data source interface scheme "{scheme}" between "{itf_a.__module__}" and "{itf_b.__module__}"',
            "uninstall one of the data source interfaces",
        )


################################################################
#                       Analytics Errors
################################################################


class AnalyticsManagerInternalError(KestrelInternalError):
    def __init__(self, error):
        super().__init__(error)


class InvalidAnalytics(KestrelException):
    def __init__(self, name, itf, msg=""):
        super().__init__(
            f'invalid analytics "{name}" at interface "{itf}". {msg}',
            "please check analytics availability",
        )


class AnalyticsError(KestrelException):
    def __init__(self, error):
        super().__init__(error, "report to analytics developer")


class AnalyticsInterfaceNotFound(KestrelException):
    def __init__(self, scheme):
        super().__init__(
            f'interface handling "{scheme}://" does not exist',
            "(re)install the missing/broken data source interface package",
        )


class InvalidAnalyticsInterfaceImplementation(Exception):
    def __init__(self, error):
        super().__init__(error, "report to analytics interface developer")


class ConflictingAnalyticsInterfaceScheme(KestrelException):
    def __init__(self, itf_a, itf_b, scheme):
        super().__init__(
            f'conflicting analytics interface scheme "{scheme}" between "{itf_a.__module__}" and "{itf_b.__module__}"',
            "uninstall one of the analytics interfaces",
        )



################################################################
#                       Module Summary
#
# - Code generation for each command in kestrel.lark
#   - The execution function names match commands in kestrel.lark
# - Each command takes 2 arguments
#     ( statement, session )
#   - statement is the current statement to process,
#     which is a dict from the parser
#   - session is the current session (context)
# - Every command returns a tuple (VarStruct, Display)
#   - VarStruct is a new object associated with the output var
#     - VarStruct associated with stmt["output"]
#     - None for some commands, e.g., DISP, SAVE, STAT
#   - Display is the data to display on the user interface
#     - a string
#     - a list of (str,str|list(str)) tuples
#     - a table that can be imported to pandas dataframe
################################################################

import functools
import logging
import itertools
from collections import OrderedDict

from kestrel.utils import remove_empty_dicts, dedup_ordered_dicts
from kestrel.exceptions import *
from kestrel.semantics import get_entity_table, get_entity_type, get_entity_len
from kestrel.symboltable import new_var
from kestrel.syntax.parser import get_all_input_var_names
from kestrel.codegen.data import load_data, load_data_file, dump_data_to_file
from kestrel.codegen.display import DisplayDataframe, DisplayDict
from kestrel.codegen.pattern import build_pattern, or_patterns
from kestrel.codegen.relations import (
    generic_relations,
    compile_generic_relation_to_pattern,
    compile_specific_relation_to_pattern,
    compile_identical_entity_search_pattern,
    compile_x_ibm_event_search_flow_in_pattern,
    compile_x_ibm_event_search_flow_out_pattern,
    are_entities_associated_with_x_ibm_event,
)

_logger = logging.getLogger(__name__)


################################################################
#                       Private Decorators
################################################################


def _default_output(func):
    # by default, create a table/view in the backend
    # using the output var name
    # in this case, the store backend can return no VarStruct
    @functools.wraps(func)
    def wrapper(stmt, session):
        ret = func(stmt, session)
        if not ret:
            var_struct = new_var(
                session.store, stmt["output"], [], stmt, session.symtable
            )
            return var_struct, None
        else:
            return ret

    return wrapper


def _guard_empty_input(func):
    @functools.wraps(func)
    def wrapper(stmt, session):
        input_len_dict = {
            v: get_entity_len(v, session.symtable)
            for v in get_all_input_var_names(stmt)
        }
        for v, size in input_len_dict.items():
            if size == 0:
                raise EmptyInputVariable(v)
        else:
            return func(stmt, session)

    return wrapper


def _debug_logger(func):
    @functools.wraps(func)
    def wrapper(stmt, session):
        _logger.debug(f"Executing '{func.__name__}' with statement: {stmt}")
        return func(stmt, session)

    return wrapper


################################################################
#                 Code Generation for Commands
################################################################


@_debug_logger
@_default_output
def merge(stmt, session):
    entity_types = list(
        set(
            [get_entity_type(var_name, session.symtable) for var_name in stmt["inputs"]]
        )
    )
    if len(entity_types) > 1:
        raise NonUniformEntityType(entity_types)
    entity_tables = [
        get_entity_table(var_name, session.symtable) for var_name in stmt["inputs"]
    ]
    session.store.merge(stmt["output"], entity_tables)
    output = new_var(session.store, stmt["output"], [], stmt, session.symtable)
    return output, None


@_debug_logger
@_default_output
def new(stmt, session):
    stmt["type"] = load_data(session.store, stmt["output"], stmt["data"], stmt["type"])


@_debug_logger
@_default_output
def load(stmt, session):
    stmt["type"] = load_data_file(
        session.store, stmt["output"], stmt["path"], stmt["type"]
    )


@_debug_logger
@_guard_empty_input
def save(stmt, session):
    dump_data_to_file(
        session.store, get_entity_table(stmt["input"], session.symtable), stmt["path"]
    )
    return None, None


@_debug_logger
def info(stmt, session):
    header = session.store.columns(get_entity_table(stmt["input"], session.symtable))
    direct_attrs, associ_attrs, custom_attrs, references = [], [], [], []
    for field in header:
        if field.startswith("x_"):
            custom_attrs.append(field)
        elif (
            field.endswith("_ref")
            or field.endswith("_refs")
            or field.endswith("_reference")
            or field.endswith("_references")
        ):
            # not useful in existing version, so do not display
            references.append(field)
        elif "_ref." in field or "_ref_" in field:
            associ_attrs.append(field)
        else:
            direct_attrs.append(field)

    disp = OrderedDict()
    disp["Entity Type"] = session.symtable[stmt["input"]].type
    disp["Number of Entities"] = str(len(session.symtable[stmt["input"]]))
    disp["Number of Records"] = str(session.symtable[stmt["input"]].records_count)
    disp["Entity Attributes"] = ", ".join(direct_attrs)
    disp["Indirect Attributes"] = [
        ", ".join(g)
        for _, g in itertools.groupby(associ_attrs, lambda x: x.rsplit(".", 1)[0])
    ]
    disp["Customized Attributes"] = ", ".join(custom_attrs)
    disp["Birth Command"] = session.symtable[stmt["input"]].birth_statement["command"]
    disp["Associated Datasource"] = session.symtable[stmt["input"]].data_source
    disp["Dependent Variables"] = ", ".join(
        session.symtable[stmt["input"]].dependent_variables
    )

    return None, DisplayDict(disp)


@_debug_logger
def disp(stmt, session):
    if len(session.symtable[stmt["input"]]) > 0:
        content = session.store.lookup(
            get_entity_table(stmt["input"], session.symtable),
            stmt["attrs"],
            stmt["limit"],
        )
    else:
        content = []
    return None, DisplayDataframe(dedup_ordered_dicts(remove_empty_dicts(content)))


@_debug_logger
@_default_output
def get(stmt, session):
    local_var_name = stmt["output"] + "_local"
    return_var_name = stmt["output"]
    return_type = stmt["type"]
    start_offset = session.config["stixquery"]["timerange_start_offset"]
    end_offset = session.config["stixquery"]["timerange_stop_offset"]

    pattern = build_pattern(
        stmt["patternbody"],
        stmt["timerange"],
        start_offset,
        end_offset,
        session.symtable,
        session.store,
    )

    if "variablesource" in stmt:
        session.store.filter(
            stmt["output"],
            stmt["type"],
            get_entity_table(stmt["variablesource"], session.symtable),
            pattern,
        )
        output = new_var(session.store, return_var_name, [], stmt, session.symtable)
        _logger.debug(f"get from variable source \"{stmt['variablesource']}\"")

    elif "datasource" in stmt:
        # rs: RetStruct
        rs = session.data_source_manager.query(
            stmt["datasource"], pattern, session.session_id
        )
        query_id = rs.load_to_store(session.store)
        session.store.extract(local_var_name, return_type, query_id, pattern)
        _output = new_var(session.store, local_var_name, [], stmt, session.symtable)

        output = _output

        if session.config["prefetch"]["get"] and len(_output):
            prefetch_ret_var_name = return_var_name + "_prefetch"
            prefetch_ret_entity_table = _prefetch(
                return_type,
                prefetch_ret_var_name,
                local_var_name,
                stmt["timerange"],
                start_offset,
                end_offset,
                {local_var_name: _output},
                session.store,
                session.session_id,
                session.data_source_manager,
            )
        else:
            prefetch_ret_entity_table = None

        if prefetch_ret_entity_table:
            session.store.merge(
                return_var_name, [local_var_name, prefetch_ret_entity_table]
            )
            output = new_var(session.store, return_var_name, [], stmt, session.symtable)
        else:
            output = new_var(session.store, local_var_name, [], stmt, session.symtable)

    else:
        raise KestrelInternalError(f"unknown type of source in {str(stmt)}")

    return output, None


@_debug_logger
@_default_output
@_guard_empty_input
def find(stmt, session):
    return_type = stmt["type"]
    input_type = session.symtable[stmt["input"]].type
    input_var_name = stmt["input"]
    return_var_name = stmt["output"]
    local_var_name = stmt["output"] + "_local"
    local_var_event_name = stmt["output"] + "_asso_event"
    relation = stmt["relation"]
    is_reversed = stmt["reversed"]
    time_range = stmt["timerange"]
    event_type = "x-oca-event"
    start_offset = session.config["stixquery"]["timerange_start_offset"]
    end_offset = session.config["stixquery"]["timerange_stop_offset"]

    if return_type not in session.store.tables():
        # return empty variable
        output = new_var(session.store, None, [], stmt, session.symtable)

    else:
        _symtable = {input_var_name: session.symtable[input_var_name]}

        event_pattern = None

        # First, get information from local store
        if relation in generic_relations:
            raw_pattern_body = compile_generic_relation_to_pattern(
                return_type, input_type, input_var_name
            )

            if (
                event_type in session.store.tables()
                and are_entities_associated_with_x_ibm_event([input_type, return_type])
                and input_type != return_type
            ):
                try:
                    event_in_pattern_body = compile_x_ibm_event_search_flow_in_pattern(
                        input_type, input_var_name
                    )
                    event_in_pattern = build_pattern(
                        event_in_pattern_body,
                        time_range,
                        start_offset,
                        end_offset,
                        _symtable,
                        session.store,
                    )
                    session.store.extract(
                        local_var_event_name, event_type, None, event_in_pattern
                    )
                    _symtable[local_var_event_name] = new_var(
                        session.store, local_var_event_name, [], stmt, session.symtable
                    )

                    event_out_pattern_body = (
                        compile_x_ibm_event_search_flow_out_pattern(
                            return_type, local_var_event_name
                        )
                    )
                    event_pattern = build_pattern(
                        event_out_pattern_body,
                        time_range,
                        start_offset,
                        end_offset,
                        _symtable,
                        session.store,
                    )
                except InvalidAttribute:
                    _logger.warning(
                        "attributes not in DB when building event pattern for x-oca-event"
                    )
        else:
            raw_pattern_body = compile_specific_relation_to_pattern(
                return_type, relation, input_type, is_reversed, input_var_name
            )

        try:
            local_pattern = build_pattern(
                raw_pattern_body,
                time_range,
                start_offset,
                end_offset,
                _symtable,
                session.store,
            )
        except InvalidAttribute:
            local_pattern = None

        local_pattern = or_patterns([local_pattern, event_pattern])
        if not local_pattern:
            _logger.info(f'no relation "{relation}" on this dataset')

        # by default, `session.store.extract` will generate new entity_table named `local_var_name`
        session.store.extract(local_var_name, return_type, None, local_pattern)

        _output = new_var(session.store, local_var_name, [], stmt, session.symtable)

        # default output without remote query
        output = _output

        # Second, prefetch all records of the entities and associated entities
        if session.config["prefetch"]["find"] and len(_output) and _output.data_source:
            prefetch_ret_var_name = return_var_name + "_prefetch"
            prefetch_ret_entity_table = _prefetch(
                return_type,
                prefetch_ret_var_name,
                local_var_name,
                time_range,
                start_offset,
                end_offset,
                {local_var_name: _output},
                session.store,
                session.session_id,
                session.data_source_manager,
            )
        else:
            prefetch_ret_entity_table = None

        if prefetch_ret_entity_table:
            session.store.merge(
                return_var_name, [local_var_name, prefetch_ret_entity_table]
            )
            output = new_var(session.store, return_var_name, [], stmt, session.symtable)
        else:
            output = new_var(session.store, local_var_name, [], stmt, session.symtable)

    return output, None


@_debug_logger
@_default_output
@_guard_empty_input
def join(stmt, session):
    session.store.join(
        stmt["output"],
        get_entity_table(stmt["input"], session.symtable),
        stmt["path"],
        get_entity_table(stmt["input_2"], session.symtable),
        stmt["path_2"],
    )


@_debug_logger
@_default_output
@_guard_empty_input
def group(stmt, session):
    session.store.assign(
        stmt["output"],
        get_entity_table(stmt["input"], session.symtable),
        op="group",
        by=stmt["path"],
    )


@_debug_logger
@_default_output
@_guard_empty_input
def sort(stmt, session):
    session.store.assign(
        stmt["output"],
        get_entity_table(stmt["input"], session.symtable),
        op="sort",
        by=stmt["path"],
        ascending=stmt["ascending"],
    )


@_debug_logger
@_default_output
@_guard_empty_input
def apply(stmt, session):
    arg_vars = [session.symtable[v_name] for v_name in stmt["inputs"]]
    display = session.analytics_manager.execute(
        stmt["workflow"], arg_vars, session.session_id, stmt["parameter"]
    )
    return None, display


################################################################
#                       Helper Functions
################################################################


def _prefetch(
    return_type,
    return_var_name,
    input_var_name,
    time_range,
    start_offset,
    end_offset,
    symtable,
    store,
    session_id,
    ds_manager,
):
    """prefetch identical entities and associated entities.

    Put the input entities in the center of an observation and query the remote
    data source of associated with input variable, so we get back:

    1. all records about the input entities.

    2. associated entities such as parent/child processes of processes, processes of network-traffic, etc.

    The function does not have explicit return, but a side effect: a view in
    the store named after `return_var_name`.

    Args:
        input_var_name (str): input variable name.

        return_var_name (str): return variable name.

        return_type (str): return entity type.

        time_range ((str, str)): start and end time in ISOTIMESTAMP.

        start_offset (int): start time offset by seconds.

        end_offset (int): end time offset by seconds.

        symtable ({str:VarStruct}): should has ``input_var_name``.

        store (firepit.SqlStorage): store.

        session_id (str): session ID.

    Returns:
        str: the entity table in store if the prefetch is performed else None.
    """

    pattern_body = compile_identical_entity_search_pattern(
        input_var_name, symtable[input_var_name]
    )

    remote_pattern = build_pattern(
        pattern_body, time_range, start_offset, end_offset, symtable, store
    )

    if remote_pattern:
        data_source = symtable[input_var_name].data_source
        resp = ds_manager.query(data_source, remote_pattern, session_id)
        query_id = resp.load_to_store(store)

        # build the return_var_name view in store
        store.extract(return_var_name, return_type, query_id, remote_pattern)

        return return_var_name

    return None



import importlib
import pkgutil
import inspect
import sys
from kestrel.analytics import AbstractAnalyticsInterface
from kestrel.exceptions import (
    AnalyticsInterfaceNotFound,
    InvalidAnalyticsInterfaceImplementation,
    ConflictingAnalyticsInterfaceScheme,
)


class AnalyticsManager:
    def __init__(self):
        self.scheme_to_interface = {}
        interfaces = _load_analytics_interfaces()
        for i in interfaces:
            self.scheme_to_interface.update({s: i for s in i.schemes()})

    def schemes(self):
        return list(self.scheme_to_interface.keys())

    def list_analytics_from_scheme(self, scheme):
        scheme = scheme.lower()
        if scheme not in self.scheme_to_interface:
            raise AnalyticsInterfaceNotFound(scheme)
        return self.scheme_to_interface[scheme].list_analytics()

    def execute(self, uri, argument_variables, session_id, parameters):
        scheme = uri.split("://")[0]
        scheme = scheme.lower()
        if scheme not in self.scheme_to_interface:
            raise AnalyticsInterfaceNotFound(scheme)
        rs = self.scheme_to_interface[scheme].execute(
            uri, argument_variables, session_id, parameters
        )
        return rs


def _list_analytics_interfaces():
    pkg_names = map(lambda x: x.name, pkgutil.iter_modules())
    itf_names = filter(lambda x: x.startswith("kestrel_analytics_"), pkg_names)
    return list(itf_names)


def _load_analytics_interfaces():
    interface_names = _list_analytics_interfaces()
    interfaces = []
    for interface_name in interface_names:
        importlib.import_module(interface_name)
        cls = inspect.getmembers(sys.modules[interface_name], _is_interface_class)
        if not cls:
            raise InvalidAnalyticsInterfaceImplementation(
                f'no interface class found in "{interface_name}"'
            )
        elif len(cls) > 1:
            raise InvalidAnalyticsInterfaceImplementation(
                f'more than one interface class found in "{interface_name}"'
            )
        else:
            interface = cls[0][1]
            interface_conflict, scheme_conflict = _find_scheme_conflict(
                interface, interfaces
            )
            if interface_conflict:
                raise ConflictingAnalyticsInterfaceScheme(
                    interface, interface_conflict, scheme_conflict
                )
            interfaces.append(interface)
    return interfaces


def _is_interface_class(obj):
    return inspect.isclass(obj) and obj.__bases__[0] == AbstractAnalyticsInterface


def _find_scheme_conflict(new_interface, interfaces):
    for interface in interfaces:
        for scheme_new in new_interface.schemes():
            for scheme_old in interface.schemes():
                if scheme_new == scheme_old:
                    return interface, scheme_new
    return None, None
